
#include "stdafx.h"

//////////////////////////////////////////////////////////////////////////////////
// Helper Macros - use as you please

#define GETFLOAT(p) *((float*)&p)
#define GETBOOL(p) *((bool*)&p)
#define GETINT(p) p
#define GETSTRING(p) *((char**)&p)
#define GETFLOATARRAY(p) p ? ((float*)p+1) : 0
#define GETINTARRAY(p) p ? ((int*)p+1) : 0
#define GETSTRINGARRAY(p) p ? ((char**)p+1) : 0
#define GETARRAYSIZE(p) p ? *((int*)p) : 0
#define GETFRAME(p) p ? ((float*)p+1) : 0
#define GETFRAMESIZE(p) p ? *((int*)p) : 0
#define GETBITMAPWIDTH(p) p ? *((int*)p) : 0
#define GETBITMAPHEIGHT(p) p ? *((int*)p+1) : 0
#define GETBITMAPCHANNELS(p) p ? *((int*)p+2) : 0
#define GETBITMAPDATA(p) p ? ((BYTE*)p+12) : 0
#define GETBITMAPBYTES(p) p ? *((int*)p) * *((int*)p+1) * *((int*)p+2) : 0
#define NEWINTARRAY(p,n) if(n>0) { *((int**)&p)=new int[n+1]; ((int*)p)[0]=n; }
#define NEWFLOATARRAY(p,n) if(n>0) { *((float**)&p)=new float[n+1]; ((int*)p)[0]=n; }
#define NEWSTRINGARRAY(p,n) if(n>0) { *((char***)&p)=new char*[n+1]; ((int*)p)[0]=n; }
#define DELETESTRING(p) if(p) { delete *((char**)&p); p=0; }
#define DELETEINTARRAY(p) if(p) { delete *((int**)&p); p=0; }
#define DELETEFLOATARRAY(p) if(p) { delete *((float**)&p); p=0; }
#define DELETESTRINGARRAY(p) if(p) { for( int j=0; j<*((int*)p); j++ ) { if( ((char**)p+1)[j] ) delete ((char**)p+1)[j]; } delete *((char***)&p); p=0; }

// Connectors for GetSysInfo
#define OUT_PAGESIZE pOut[0]
#define OUT_ALLOCSIZE pOut[1]

// Connectors for MakeMappedFile
#define IN_MAPNAME pIn[0]
#define IN_MAPSIZE pIn[1]
#define OUT_MAPADDRESS pOut[2]

// Connectors for WriteFrame
#define IN_FRAME pIn[0]

// Connectors for ReadFrame
#define IN_MIX pIn[1]
#define OUT_FRAME pOut[0]

// DLL Global variables
static LPVOID lpMem = NULL;
static HANDLE hMapObject = NULL;
static int    memSize = 0;

//////////////////////////////////////////////////////////////////////////////////

extern "C" __declspec(dllexport) void GetSysInfo( int nParams, int* pIn, int* pOut )
{
	if( pIn && pOut && nParams >= 2 )
	{
		SYSTEM_INFO siSysInfo;

		// Get Windows system information for file and memory boundaries
		GetSystemInfo(&siSysInfo);
		
		// Send to outputs
		OUT_PAGESIZE  = siSysInfo.dwPageSize;
		OUT_ALLOCSIZE = siSysInfo.dwAllocationGranularity;
	}
}


extern "C" __declspec(dllexport) void OpenMapVirtual( int nParams, int* pIn, int* pOut )
{
	if ( pIn && pOut && nParams >= 3 )
	{
		char* mapName = GETSTRING(IN_MAPNAME);
		int   nameLen = (mapName ? strlen(mapName) : 0);
		int   mapSize = IN_MAPSIZE;

		// If a mapping is already open, close it
		if ( hMapObject != NULL )
		{
			// Unmap the view
			UnmapViewOfFile(lpMem);

			// Close the mapping
			CloseHandle(hMapObject);
		}

		// Attempt to open the mapped file.  hMapObject is NULL if fails
		if ( nameLen > 0 && mapSize > 0 )
		{
			hMapObject = CreateFileMapping(
				INVALID_HANDLE_VALUE,
				NULL,
				PAGE_READWRITE,
				0,
				mapSize,
				(LPCWSTR)mapName);

		}
		else
		{
			hMapObject = NULL;
		}

		// Check if this was the first time the map was opened
		bool firstOpen = ( GetLastError() != ERROR_ALREADY_EXISTS);

		if ( hMapObject != NULL)
		{
			// There is a mapping, so create a view into it
			lpMem = MapViewOfFile(
				hMapObject,
				FILE_MAP_WRITE,
				0,
				0,
				0);
		}
		else
		{
			// No map, so NULL the view pointer
			lpMem = NULL;
		}

		if ( lpMem != NULL)
		{
			// A view was created

			// If newly opened, zero the memory
			if ( firstOpen )
			{
				memset(lpMem, 0, mapSize);
			}

			// And send the address to the output
			memSize = mapSize;
			OUT_MAPADDRESS = (int)lpMem;
		}
		else
		{
			memSize = 0;
			OUT_MAPADDRESS = 0;
		}

	}
}


extern "C" __declspec(dllexport) void CloseMap( int nParams, int* pIn, int* pOut )
{
	if ( hMapObject != NULL )
	{
		// Unmap the view
		UnmapViewOfFile(lpMem);

		// Close the mapping
		CloseHandle(hMapObject);
	}
}


extern "C" __declspec(dllexport) void WriteFrame( int nParams, int* pIn, int* pOut )
{
	if ( pIn && pOut && nParams >= 1 )
	{
		if ( IN_FRAME && lpMem != NULL )
		{         
			int size = min( memSize, GETFRAMESIZE(IN_FRAME) );

			if ( size > 0 )
			{
				float* frameData = GETFRAME(IN_FRAME);
				float*   memData = (float*)lpMem;

				for( int i = 0 ; i < size ; i++ )
				{
				  memData[i] = frameData[i];
				}
			}
		}
	}
}


extern "C" __declspec(dllexport) void ReadFrame( int nParams, int* pIn, int* pOut )
{
	if ( pIn && pOut && nParams >= 2 )
	{
		if ( OUT_FRAME && lpMem != NULL )
		{         
			int size = min( memSize, GETFRAMESIZE(OUT_FRAME) );

			if ( size > 0 )
			{
				float* frameData = GETFRAME(OUT_FRAME);
				float*   memData = (float*)lpMem;
				bool         mix = GETBOOL(IN_MIX);
				
				if ( mix )
				{
					for( int i = 0 ; i < size ; i++ )
					{
					  frameData[i] = memData[i] + frameData[i];
					}
				}
				else
				{
					for( int i = 0 ; i < size ; i++ )
					{
					  frameData[i] = memData[i];
					}
				}
			}
		}
	}
}
